# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: changelog_schema.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x63hangelog_schema.proto\"\xb0\x01\n\x14\x63hangelog_schema_pb2\x12\x11\n\tCLIENT_ID\x18\x01 \x02(\t\x12\x11\n\tBUCKET_ID\x18\x02 \x02(\t\x12\x11\n\tFILE_PATH\x18\x03 \x02(\t\x12\x11\n\tFILE_NAME\x18\x04 \x01(\t\x12\x0e\n\x06\x41\x43TION\x18\x05 \x01(\t\x12\x11\n\tACTION_AT\x18\x06 \x01(\t\x12\x12\n\nREPLICATED\x18\x07 \x01(\x08\x12\x15\n\rREPLICATED_AT\x18\x08 \x01(\t')



_CHANGELOG_SCHEMA_PB2 = DESCRIPTOR.message_types_by_name['changelog_schema_pb2']
changelog_schema_pb2 = _reflection.GeneratedProtocolMessageType('changelog_schema_pb2', (_message.Message,), {
  'DESCRIPTOR' : _CHANGELOG_SCHEMA_PB2,
  '__module__' : 'changelog_schema_pb2'
  # @@protoc_insertion_point(class_scope:changelog_schema_pb2)
  })
_sym_db.RegisterMessage(changelog_schema_pb2)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CHANGELOG_SCHEMA_PB2._serialized_start=27
  _CHANGELOG_SCHEMA_PB2._serialized_end=203
# @@protoc_insertion_point(module_scope)
