TIME_REGEX = "T(2[0-3]|[01][0-9])([0-5][0-9])([0-5][0-9])Z"
DATE_REGEX = "([1-2][0-9][0-9][0-9])(0[1-9]|1[0-2])(0[1-9]|[1-2][0-9]|3[01])"

SA_SUFFIX = '@indx-data-services.iam.gserviceaccount.com'

DATASET_OWNER = 'data-lake-generator@indx-data-services.iam.gserviceaccount.com'

SIZE_OF_FILES = 0.0
